#include <QTime>
#include <QDebug>
#include "Map.h"

CMap  map;

//---------------------------------------------------------------------------------------------
CWall::CWall()
{
    mType = 0;
}

//---------------------------------------------------------------------------------------------
CWall::~CWall()
{
    deleteParams();
}

//---------------------------------------------------------------------------------------------
void CWall::deleteParams()
{
    for (size_t i = 0; i < mParams.size(); ++i)
        delete mParams[i];
    mParams.clear();
}

//---------------------------------------------------------------------------------------------

#define NEW_PARAM(_enum,_struct) \
    case _enum: \
        newParam = (CParam*)(new _struct); \
        break;

void CWall::setType(uint8_t type)
{
    if (mType == type)
        return;

    mType = type;

    // allocate the params
    deleteParams();
    std::vector<CParamType>& paramTypes = map.mWallsParams[type];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        CParam* newParam = NULL;

        switch (paramTypes[i].mType)
        {
            NEW_PARAM(eParamOrnate,CParamOrnate)
            NEW_PARAM(eParamInt,CParamInt)
            NEW_PARAM(eParamBool,CParamBool)

            default:
                break;
        }
        mParams.push_back(newParam);
    }
}

//---------------------------------------------------------------------------------------------
uint8_t CWall::getType()
{
    return  mType;
}

//---------------------------------------------------------------------------------------------
// assignment operator used for clipboard functions and copy/cut/paste tools

#define COPY_PARAM(_enum,_struct) \
    case _enum: \
    { \
        _struct*   lParam = (_struct*)mParams[i]; \
        _struct*   rParam = (_struct*)rhs.mParams[i]; \
        lParam->mValue = rParam->mValue; \
    } \
    break;

CWall&  CWall::operator=(CWall& rhs)
{
    // copy the type and alloc the params
    setType(rhs.mType);

    // copy the params
    for (size_t i = 0; i < mParams.size(); ++i)
    {
        switch (mParams[i]->mType)
        {
            COPY_PARAM(eParamOrnate,CParamOrnate)
            COPY_PARAM(eParamBool,CParamBool)
            COPY_PARAM(eParamInt,CParamInt)

            default:
                break;
        }
    }
    return *this;
}

//---------------------------------------------------------------------------------------------

#define LOAD_PARAM(_enum,_struct,_type,_expr) \
    case _enum: \
        { \
            _type value; \
            fread(&value, sizeof(value), 1, handle); \
            ((_struct*)mParams[i])->mValue = (_expr); \
        } \
        break;

#define LOAD_STRING_PARAM(_enum,_struct) \
    case _enum: \
        { \
            static char   temp[256]; \
            uint8_t size; \
            fread(&size, 1, 1, handle); \
            fread(temp, size, 1, handle); \
            temp[size] = 0; \
            ((_struct*)mParams[i])->mValue = temp; \
        } \
        break;

void  CWall::load(FILE* handle)
{
    uint8_t type;
    fread(&type, 1, 1, handle);
    setType(type);

    std::vector<CParamType>& paramTypes = map.mWallsParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            LOAD_PARAM(eParamOrnate, CParamOrnate, uint8_t, value)
            LOAD_PARAM(eParamInt, CParamInt, int32_t, value)
            LOAD_PARAM(eParamBool, CParamBool, uint8_t, (value == 0 ? false : true))

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------

#define SAVE_PARAM(_enum,_struct,_type,_origType,_expr) \
case _enum: \
    { \
        _type value; \
\
        if (i < mParams.size()) \
        { \
            _origType    temp = ((_struct*)mParams[i])->mValue; \
            value = _expr; \
        } \
        else \
            value = 0; \
\
        fwrite(&value, sizeof(value), 1, handle); \
    } \
    break;

#define SAVE_STRING_PARAM(_enum,_struct) \
case _enum: \
    { \
        std::string value; \
\
        if (i < mParams.size()) \
        { \
            value = ((_struct*)mParams[i])->mValue; \
        } \
\
        uint8_t size = value.size(); \
        fwrite(&size, 1, 1, handle); \
        fwrite(value.c_str(), size, 1, handle); \
    } \
    break;

void  CWall::save(FILE* handle)
{
    fwrite(&mType, 1, 1, handle);

    std::vector<CParamType>& paramTypes = map.mWallsParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            SAVE_PARAM(eParamOrnate, CParamOrnate, uint8_t, uint8_t, temp)
            SAVE_PARAM(eParamInt, CParamInt, int32_t, int32_t, temp)
            SAVE_PARAM(eParamBool, CParamBool, uint8_t, bool, (temp == false ? 0 : 1))

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
CParam*    CWall::findParamByName(QString name)
{
    for (size_t i = 0; i < map.mWallsParams[mType].size(); ++i)
    {
        if (map.mWallsParams[mType][i].mName == name)
            return mParams[i];
    }

    return NULL;
}

//---------------------------------------------------------------------------------------------
uint8_t CWall::getOrnateParam(QString name)
{
    CParamOrnate* param = (CParamOrnate*)findParamByName(name);

    if (param != NULL && param->mType == eParamOrnate)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
int32_t CWall::getIntParam(QString name)
{
    CParamInt* param = (CParamInt*)findParamByName(name);

    if (param != NULL && param->mType == eParamInt)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
bool    CWall::getBoolParam(QString name)
{
    CParamBool* param = (CParamBool*)findParamByName(name);

    if (param != NULL && param->mType == eParamBool)
        return param->mValue;
    else
        return false;
}

//---------------------------------------------------------------------------------------------
void    CWall::setBoolParam(QString name, bool value)
{
    CParamBool* param = (CParamBool*)findParamByName(name);

    if (param != NULL && param->mType == eParamBool)
        param->mValue = value;
}

//---------------------------------------------------------------------------------------------
CTile::CTile()
{
    mType = 0;
}

//---------------------------------------------------------------------------------------------
CTile::~CTile()
{
    deleteParams();
}

//---------------------------------------------------------------------------------------------
void CTile::deleteParams()
{
    for (size_t i = 0; i < mParams.size(); ++i)
        delete mParams[i];
    mParams.clear();
}

//---------------------------------------------------------------------------------------------

#define ALLOC_PARAM(_enum,_struct) \
    case _enum: \
    { \
        _struct*   lParam = (_struct*)mParams[i]; \
        _struct*   rParam = (_struct*)rhs.mParams[i]; \
        lParam->mValue = rParam->mValue; \
    } \
    break;

void CTile::setType(uint8_t type)
{
    if (mType == type)
        return;

    mType = type;

    // allocate the params
    deleteParams();
    std::vector<CParamType>& paramTypes = map.mTilesParams[type];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        CParam* newParam = NULL;

        switch (paramTypes[i].mType)
        {
            NEW_PARAM(eParamOrnate,CParamOrnate)
            NEW_PARAM(eParamDoor,CParamDoor)
            NEW_PARAM(eParamDoorOrnate,CParamDoorOrnate)
            NEW_PARAM(eParamEnum,CParamEnum)
            NEW_PARAM(eParamInt,CParamInt)
            NEW_PARAM(eParamBool,CParamBool)
            NEW_PARAM(eParamScript,CParamScript)

            default:
                break;
        }
        mParams.push_back(newParam);
    }
}

//---------------------------------------------------------------------------------------------
uint8_t CTile::getType()
{
    return  mType;
}

//---------------------------------------------------------------------------------------------
// copy the tile's type and it's params, but not the walls
void  CTile::copyWithoutWalls(CTile& rhs)
{
    // copy the type and alloc the params
    setType(rhs.mType);

    // copy the params
    for (size_t i = 0; i < mParams.size(); ++i)
    {
        switch (mParams[i]->mType)
        {
            COPY_PARAM(eParamOrnate,CParamOrnate)
            COPY_PARAM(eParamDoor,CParamDoor)
            COPY_PARAM(eParamDoorOrnate,CParamDoorOrnate)
            COPY_PARAM(eParamEnum,CParamEnum)
            COPY_PARAM(eParamBool,CParamBool)
            COPY_PARAM(eParamInt,CParamInt)
            COPY_PARAM(eParamScript,CParamScript)

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
// assignment operator used for clipboard functions and copy/cut/paste tools
CTile&  CTile::operator=(CTile& rhs)
{
    copyWithoutWalls(rhs);

    // copy the walls
    for (int i = 0; i < eWallSideMax; ++i)
        mWalls[i] = rhs.mWalls[i];

    return *this;
}

//---------------------------------------------------------------------------------------------
CParam* CTile::findParamByName(QString name)
{
    for (size_t i = 0; i < map.mTilesParams[mType].size(); ++i)
    {
        if (map.mTilesParams[mType][i].mName == name)
            return mParams[i];
    }

    return NULL;
}

//---------------------------------------------------------------------------------------------
uint8_t CTile::getOrnateParam(QString name)
{
    CParamOrnate* param = (CParamOrnate*)findParamByName(name);

    if (param != NULL && param->mType == eParamOrnate)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
uint8_t CTile::getEnumParam(QString name)
{
    CParamEnum* param = (CParamEnum*)findParamByName(name);

    if (param != NULL && param->mType == eParamEnum)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
uint8_t CTile::getDoorParam(QString name)
{
    CParamDoor* param = (CParamDoor*)findParamByName(name);

    if (param != NULL && param->mType == eParamDoor)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
int32_t CTile::getIntParam(QString name)
{
    CParamInt* param = (CParamInt*)findParamByName(name);

    if (param != NULL && param->mType == eParamInt)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
void CTile::setIntParam(QString name, int32_t value)
{
    CParamInt* param = (CParamInt*)findParamByName(name);

    if (param != NULL && param->mType == eParamInt)
        param->mValue = value;
}

//---------------------------------------------------------------------------------------------
bool    CTile::getBoolParam(QString name)
{
    CParamBool* param = (CParamBool*)findParamByName(name);

    if (param != NULL && param->mType == eParamBool)
        return param->mValue;
    else
        return false;
}

//---------------------------------------------------------------------------------------------
void    CTile::setBoolParam(QString name, bool value)
{
    CParamBool* param = (CParamBool*)findParamByName(name);

    if (param != NULL && param->mType == eParamBool)
        param->mValue = value;
}

//---------------------------------------------------------------------------------------------
uint8_t CTile::getDoorOrnateParam(QString name)
{
    CParamDoorOrnate* param = (CParamDoorOrnate*)findParamByName(name);

    if (param != NULL && param->mType == eParamDoorOrnate)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
std::string CTile::getScriptParam(QString name)
{
    CParamScript* param = (CParamScript*)findParamByName(name);

    if (param != NULL && param->mType == eParamScript)
        return param->mValue;
    else
        return 0;
}

//---------------------------------------------------------------------------------------------
void  CTile::load(FILE* handle)
{
    uint8_t type;
    fread(&type, 1, 1, handle);
    setType(type);

    std::vector<CParamType>& paramTypes = map.mTilesParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            LOAD_PARAM(eParamOrnate, CParamOrnate, uint8_t, value)
            LOAD_PARAM(eParamDoor, CParamDoor, uint8_t, value)
            LOAD_PARAM(eParamDoorOrnate, CParamDoorOrnate, uint8_t, value)
            LOAD_PARAM(eParamEnum, CParamEnum, uint8_t, value)
            LOAD_PARAM(eParamInt, CParamInt, int32_t, value)
            LOAD_PARAM(eParamBool, CParamBool, uint8_t, (value == 0 ? false : true))
            LOAD_STRING_PARAM(eParamScript, CParamScript)

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
void  CTile::save(FILE* handle)
{
    fwrite(&mType, 1, 1, handle);

    std::vector<CParamType>& paramTypes = map.mTilesParams[mType];

    for (size_t i = 0; i < paramTypes.size(); ++i)
    {
        switch (paramTypes[i].mType)
        {
            SAVE_PARAM(eParamOrnate, CParamOrnate, uint8_t, uint8_t, temp)
            SAVE_PARAM(eParamDoor, CParamDoor, uint8_t, uint8_t, temp)
            SAVE_PARAM(eParamDoorOrnate, CParamDoorOrnate, uint8_t, uint8_t, temp)
            SAVE_PARAM(eParamEnum, CParamEnum, uint8_t, uint8_t, temp)
            SAVE_PARAM(eParamInt, CParamInt, int32_t, int32_t, temp)
            SAVE_PARAM(eParamBool, CParamBool, uint8_t, bool, (temp == false ? 0 : 1))
            SAVE_STRING_PARAM(eParamScript, CParamScript)

            default:
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
CMap::CMap()
{
    // default values
    mSize = CVec2(15, 10);
    mTiles = NULL;
}

//---------------------------------------------------------------------------------------------
CMap::~CMap()
{
    if (mTiles != NULL)
        delete[] mTiles;

    mWallsParams.clear();
}

//---------------------------------------------------------------------------------------------
void CMap::resize()
{
    if (mTiles != NULL)
        delete[] mTiles;

    mTiles = new CTile[mSize.x * mSize.y];
}

//---------------------------------------------------------------------------------------------
void   CMap::setTile(CVec2 pos, uint8_t type)
{
    if (pos.x >= 0 && pos.x < mSize.x &&
        pos.y >= 0 && pos.y < mSize.y)
    {
        getTile(pos)->setType(type);
    }
}

//---------------------------------------------------------------------------------------------
void   CMap::setWall(CVec2 pos, EWallSide side, uint8_t type)
{
    if (pos.x >= 0 && pos.x < mSize.x &&
        pos.y >= 0 && pos.y < mSize.y)
    {
        getTile(pos)->mWalls[side].setType(type);
    }
}

//---------------------------------------------------------------------------------------------
CTile*  CMap::getTile(CVec2 pos)
{
    if (pos.x < 0 || pos.x >= mSize.x ||
        pos.y < 0 || pos.y >= mSize.y)
        return NULL;

    return(&mTiles[pos.y * mSize.x + pos.x]);
}

//---------------------------------------------------------------------------------------------
bool CMap::isLoaded()
{
    return mTiles != NULL;
}

//---------------------------------------------------------------------------------------------
char    gFileHeader[] = "DMMP"; // Dungeon Master MaP
#define CURRENT_VERSION     1
#define CURRENT_REVISION    3

void CMap::save(char* fileName)
{
    FILE*       handle = fopen(fileName, "wb");
    uint16_t    ver = CURRENT_VERSION;
    uint16_t    rev = CURRENT_REVISION;

    if (handle != NULL)
    {
        // header, version, revision
        fwrite(gFileHeader, 1, 4, handle);
        fwrite(&ver, 2, 1, handle);
        fwrite(&rev, 2, 1, handle);

        // map size
        fwrite(&mSize.x, 4, 1, handle);
        fwrite(&mSize.y, 4, 1, handle);

        // map data
        for (int y = 0; y < mSize.y; ++y)
            for (int x = 0; x < mSize.x; ++x)
            {
                CTile*  t = getTile(CVec2(x, y));
                t->save(handle);

                for (int side = 0; side < eWallSideMax; ++side)
                    t->mWalls[side].save(handle);
            }

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
void CMap::load(char* fileName)
{
    FILE*       handle = fopen(fileName, "rb");
    char        header[4 + 1];
    uint16_t    ver;
    uint16_t    rev;

    if (handle != NULL)
    {
        // header, version, revision
        fread(header, 1, 4, handle);
        header[4] = 0;
        fread(&ver, 2, 1, handle);
        fread(&rev, 2, 1, handle);

#ifndef CONVERT_MAP
        // check if it's the right version
        if (strcmp(header, gFileHeader) != 0 ||
            ver != CURRENT_VERSION ||
            rev != CURRENT_REVISION)
        {
            fclose(handle);
            return;
        }
#endif

        // map size
        fread(&mSize.x, 4, 1, handle);
        fread(&mSize.y, 4, 1, handle);

        // allocate tiles
        if (mTiles != NULL)
            delete[] mTiles;

        mTiles = new CTile[mSize.x * mSize.y];

        // map data
        for (int y = 0; y < mSize.y; ++y)
            for (int x = 0; x < mSize.x; ++x)
            {
                CTile*  t = getTile(CVec2(x, y));
                t->load(handle);

                for (int side = 0; side < eWallSideMax; ++side)
                    t->mWalls[side].load(handle);
            }

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
CParamOrnate::CParamOrnate()
{
    mType = eParamOrnate;
    mValue = 0;
}

//---------------------------------------------------------------------------------------------
CParamDoor::CParamDoor()
{
    mType = eParamDoor;
    mValue = 0;
}

//---------------------------------------------------------------------------------------------
CParamDoorOrnate::CParamDoorOrnate()
{
    mType = eParamDoorOrnate;
    mValue = 0;
}

//---------------------------------------------------------------------------------------------
CParamEnum::CParamEnum()
{
    mType = eParamEnum;
    mValue = 0;
}

//---------------------------------------------------------------------------------------------
CParamInt::CParamInt()
{
    mType = eParamInt;
    mValue = 0;
}

//---------------------------------------------------------------------------------------------
CParamBool::CParamBool()
{
    mType = eParamBool;
    mValue = false;
}

//---------------------------------------------------------------------------------------------
CParamScript::CParamScript()
{
    mType = eParamScript;
    mValue = "";
}
